#include <stddef.h>
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "debug.h"
#include "except.h"
#include "bool.h"
#include "aa.h"
#include "vector.h"
#include "matrix.h"
#include "doublematrix.h"
#include "block.h"
#include "weighting.h"
#include "counts.h"
#include "minrisk.h"
#include "blosum.h"
#include "pssm.h"
#include "main.h"

/************************************************************************
 *
 ************************************************************************/

/*
static void
print_distribution_html (double **m) {
  int i;

  for (i = 0; i < NAT_AAS; ++i) {
    printf("<PARAM NAME=%c VALUE=\"%g\">\n",protein1_aa(i),m[0][i]);
  }
}
*/

static Vector_T
get_background_new () {
  int i;
  double qi;
  Vector_T background = Vector_new(NAT_AAS);

  for (i = 0; i < NAT_AAS; ++i) {
    qi = AA_acd_freq(i);
    Vector_put(background,i,qi);
  }
  return background;
}

static Vector_T
normalize_new (Vector_T counts, double N) {  
  int i;
  double p;
  Vector_T freqs = Vector_new(Vector_length(counts));
  
  for (i = 0; i < Vector_length(counts); ++i) {
    p = Vector_get(counts,i);
    p /= N;
    Vector_put(freqs,i,p);
  }
  return freqs;
}

static Vector_T
lambda_subst_new (Vector_T phat, double **cond) {
  int j, k, residues;
  double lambdaj;
  Vector_T lambda;

  residues = Vector_length(phat);
  lambda = Vector_new(residues);
  for (j = 0; j < residues; ++j) {
    lambdaj = 0.0;
    for (k = 0; k < residues; ++k) {
      lambdaj += Vector_get(phat,k) * cond[j][k];
    }
    Vector_put(lambda,j,lambdaj);
  }
  return lambda;
}


static Matrix_T
compute_pest_matrix_new (Matrix_T count_matrix, Vector_T background,
			 double **cond, Mode_T mode, Metric_T metric,
			 bool iterate_p) {
  int pos, positions;
  Vector_T counts, phat, pest, lambda;
  double N;
  Matrix_T pest_matrix;

  positions = Matrix_width(count_matrix);
  pest_matrix = Matrix_new(positions,NAT_AAS);
  for (pos = 0; pos < positions; ++pos) {
    counts = Matrix_get(count_matrix,pos);
    N = Vector_sum(counts);
    phat = normalize_new(counts,N);
    if (mode == BACKGROUND) {
      if (metric == SQERR) {
	if (iterate_p) {
	  pest = Minrisk_sqerr_background_iter(phat,background,cond,N);
	} else {
	  pest = Minrisk_sqerr_background_new(phat,background,cond,N);
	}
      } else {
	if (iterate_p) {
	  pest = Minrisk_relent_background_iter(phat,background,cond,N);
	} else {
	  pest = Minrisk_relent_background_new(phat,background,cond,N);
	}	  
      }
    } else {
      lambda = lambda_subst_new(phat,cond);
      if (metric == SQERR) {
	if (iterate_p) {
	  pest = Minrisk_sqerr_subst_iter(phat,lambda,cond,N);
	} else {
	  pest = Minrisk_sqerr_subst_new(phat,lambda,cond,N);
	}
      } else {
	if (iterate_p) {
	  pest = Minrisk_relent_subst_new(phat,lambda,cond,N);
	} else {
	  pest = Minrisk_relent_subst_new(phat,lambda,cond,N);
	}
      }
      Vector_free(&lambda);
    }
    Vector_free(&phat);
    Matrix_put(pest_matrix,pos,pest);
  }
  return pest_matrix;
}

static void
parse_arguments (char **blockfile, char **blosumfile, Weighting_T *weighting, 
		 Mode_T *mode, Metric_T *metric, bool *iterate_p, bool *html_p, 
                 int argc, char *argv[]) {
  ++argv; --argc;
  while ((*argv != NULL) && (*argv)[0] == '-') {
    if ((*argv)[1] == 'U') {
      *weighting = UNIFORM; argc -= 1; ++argv;
    } else if ((*argv)[1] == 'W') {
      *weighting = HENIKOFF; argc -= 1; ++argv;
    } else if ((*argv)[1] == 'B') {
      *mode = BACKGROUND; argc -= 1; ++argv;
    } else if ((*argv)[1] == 'S') {
      *mode = SUBST; ++argv; *blosumfile = *argv; argc -= 2; ++argv;
    } else if ((*argv)[1] == 'Q') {
      *metric = SQERR; argc -= 1; ++argv;
    } else if ((*argv)[1] == 'R') {
      *metric = RELENT; argc -= 1; ++argv;
    } else if ((*argv)[1] == 'I') {
      *iterate_p = TRUE; argc -= 1; ++argv;
    } else if ((*argv)[1] == 'h') {
      *html_p = TRUE; argc -= 1; ++argv;
    } else {
      fprintf(stderr,"I don't recognize the flag: %s\n",*argv);
      exit(9);
    }
  }
  if (argc > 0) {
    *blockfile = *argv;
    argc -= 1; ++argv;
  }
}

int
main (int argc, char *argv[]) {
  FILE *fp;
  char *blockfile = NULL, *blosumfile = DEFAULT_BLOSUM_FILE,
    *data_dir = DATA_DIR, *blosum_fullfile = NULL;
  Weighting_T weighting = HENIKOFF;
  Mode_T mode = SUBST;
  Metric_T metric = SQERR;
  bool iterate_p = FALSE, html_p = FALSE;
  double **joint = NULL, **cond = NULL;
  int positions, len;
  Vector_T background, weights;
  Matrix_T count_matrix, pest_matrix, logodds, scaled;
  Block_T block;

  joint = Doublematrix_new(NAT_AAS,NAT_AAS); 
  cond = Doublematrix_new(NAT_AAS,NAT_AAS); 

  TRY

  parse_arguments(&blockfile,&blosumfile,&weighting,&mode,&metric,
		  &iterate_p,&html_p,argc,argv);

  if (mode == SUBST) {
    if (strchr(blosumfile,'/')) {
      blosum_fullfile = blosumfile;
    } else {
      blosum_fullfile = (char *) calloc(200,sizeof(char));
      strcpy(blosum_fullfile,data_dir);
      len = strlen(data_dir);
      if (data_dir[len-1] != '/') {
	strcat(blosum_fullfile,"/");
      }
      strcat(blosum_fullfile,blosumfile);
    }

    Blosum_read(joint,cond,blosum_fullfile);
  }
  background = get_background_new(); 
  
  if (blockfile == NULL) {
    scanf("%d",&positions);
    block = Block_read_new(stdin,positions);
  } else {
    fp = fopen(blockfile,"r");
    if (!fp) {
      printf("Can't open block file %s\n",blockfile);
      fclose(fp);
      exit(1);
    } else {
      fscanf(fp,"%d",&positions);   
      block = Block_read_new(fp,positions);
      fclose(fp);
    }
  }

  switch (weighting) {
    case HENIKOFF: weights = Weighting_henikoff_new(block); break;
    case UNIFORM: weights = Weighting_uniform_new(block); break;
    default: weights = Weighting_uniform_new(block); break;
  }
  count_matrix = (Matrix_T) Counts_fromBlock_new(block,weights);
  Block_free(&block);
  Vector_free(&weights);
  pest_matrix = compute_pest_matrix_new(count_matrix,background,
					cond,mode,metric,iterate_p);
  Matrix_free(&count_matrix);

  if (mode == SUBST) {
    Doublematrix_free(cond,NAT_AAS);
  }
  Vector_free(&background); 
  logodds = PSSM_logodds_new(pest_matrix);
  scaled = PSSM_scaled_new(logodds);
  PSSM_print(scaled,positions);

  Matrix_free (&logodds);
  Matrix_free (&scaled);
  Matrix_free (&pest_matrix);

  ELSE
    fprintf(stderr,
	    "An internal error has occured from which there is "
	    "no recovery.\nPlease report this error by e-mail to "
	    "brutlag@stanford.edu:\n\n");
  RERAISE;
  END_TRY;

  Doublematrix_free(joint,NAT_AAS);

  fclose (stdin);
  fclose (stderr);
  fclose (stdout);

  return EXIT_SUCCESS; 
}

