#include "distvector.h"
#include "pattern.h"
#include <stdlib.h>


#define T Distvector_T
struct T {
  int length;
  Dist_T *dists;
  Pattern_T *patterns;
};


static int
markov_size (int order, int maxorder) {
  int size = 1, i;

  if (order < maxorder) {
    for (i = 0; i < order; ++i) {
      size *= NUM_BASES;
    }
    return size;
  } else {
    for (i = 0; i < maxorder; ++i) {
      size *= NUM_BASES;
    }
    return size;
  }
}

static int
markov_cum_size (int order, int maxorder) {
  int cum_size = 0, size = 1, i;

  if (order < maxorder) {
    for (i = 0; i < order; ++i) {
      size *= NUM_BASES;
      cum_size += size;
    }
    return cum_size;
  } else {
    for (i = 0; i < maxorder; ++i) {
      size *= NUM_BASES;
      cum_size += size;
    }
    return cum_size;
  }
}

T
Distvector_init (int maxorder) {
  T new = (T) malloc(sizeof(*new));
  
  new->length = 1;
  new->dists = (Dist_T *) calloc(new->length,sizeof(Dist_T));
  new->patterns = (Pattern_T *) calloc(new->length,sizeof(Pattern_T));
  new->dists[0] = Dist_init();
  new->patterns[0] = Pattern_new(0,0,maxorder);
  return new;
}

T
Distvector_new (int pos, int maxorder, int min_score, int max_score) {
  T new = (T) malloc(sizeof(*new));
  int i;
  
  new->length = markov_size(pos+1,maxorder);
  new->dists = (Dist_T *) calloc(new->length,sizeof(Dist_T));
  new->patterns = (Pattern_T *) calloc(new->length,sizeof(Pattern_T));
  for (i = 0; i < new->length; ++i) {
    new->dists[i] = Dist_new(min_score,max_score);
    new->patterns[i] = Pattern_new(i,pos,maxorder);
  }
  return new;
}
  
void
Distvector_free (T *distvector) {
  int i;

  for (i = 0; i < (*distvector)->length; ++i) {
    Dist_free(&((*distvector)->dists[i]));
    Pattern_free(&((*distvector)->patterns[i]));
  }
  free((*distvector)->dists);
  free((*distvector)->patterns);
  free(*distvector);
}


static int
shift_index (int index, int base) {
  return index * NUM_BASES + base;
}


void
Distvector_print (T distvector) {
  int i;

  for (i = 0; i < distvector->length; ++i) {
    Dist_print(distvector->dists[i]);
  }
  printf("----------------------------------------\n");
}
    


T
Distvector_update (T old, int pos, int maxorder, Frequencies_T frequencies,
		   int **profile, int *min_scores, int *max_scores) {
  T new;
  Dist_T old_dist, new_dist;
  int prob_base, i, j, a, prob_index, incr, score;
  double old_prob, prob;
  
  prob_base = markov_cum_size(pos,maxorder);
  new = Distvector_new(pos,maxorder,min_scores[pos],max_scores[pos]);
  for (i = 0; i < old->length; ++i) {
    old_dist = old->dists[i];
    for (a = A; a < Z; ++a) {
      prob_index = prob_base + shift_index(i,a);
      j = shift_index(i,a) % new->length;
      prob = Frequencies_get(frequencies,prob_index);
      new_dist = new->dists[j];
      incr = profile[pos][a];
      for (score = Dist_min_score(old_dist); 
	   score <= Dist_max_score(old_dist); ++score) {
	if ((old_prob = Dist_get_prob(old_dist,score)) > 0.0) {
	  /* printf("old:%d @ %d = %f -- aa:%d %f (%d) --> new:%d @ %d\n",
		 i,score,Dist_get_prob(old_dist,score),
		 a,prob,prob_index,
		 j,score+incr); */
	  Dist_add_prob(new_dist,score+incr,old_prob*prob);
	}
      }
    }
  }
  return new;
}
      

      
Dist_T
Distvector_compress (T distvector) {
  Dist_T new;
  int i, min_score, max_score;

  min_score = Dist_min_score(distvector->dists[0]);
  max_score = Dist_max_score(distvector->dists[0]);

  for (i = 1; i < distvector->length; ++i) {
    if (Dist_min_score(distvector->dists[i]) < min_score) {
      min_score = Dist_min_score(distvector->dists[i]);
    }
    if (Dist_max_score(distvector->dists[i]) < max_score) {
      max_score = Dist_max_score(distvector->dists[i]);
    }
  }
  new = Dist_new(min_score,max_score);
  for (i = 0; i < distvector->length; ++i) {
    Dist_add_dist(new,distvector->dists[i]);
  }
  return new;
}


double
Distvector_check (T distvector) {
  double prob = 0.0;
  int i;

  for (i = 0; i < distvector->length; ++i) {
    prob += Dist_check(distvector->dists[i]);
  }
  return prob;
}

